#include "STLAllocatorSample.h"
#include "STLAllocatorSample1.h"

#include "TDSTLAllocatorSample.h"

#include <iostream>
#include <vector>
#include <list>

class TDClass1
{
	int a;

public:

	TDClass1()
	{
		*gf << "TDClass1 Constructor " << (void*)this << std::endl;
	}

	~TDClass1()
	{
		*gf << "TDClass1 Destructor " << (void*)this << std::endl;
	}

	TDClass1(const TDClass1 & src) : a(src.a)
	{
		*gf << "TDClass1 Copy Constructor " << (void*)this << std::endl;
	}

	void SetA(int a)
	{
		this->a = a;
	}

	int GetA()
	{
		return a;
	}

};

void PrintAllMemory(unsigned char * d, size_t dsz)
{
	(*gf) << "memory dump" << std::endl;
	size_t i, isz;

	isz = dsz;
	for(i = 0; i < isz ; i++)
	{
		(*gf) << std::hex << std::setw(2) << std::setfill('0') << (int)d[i];
		if(i % 64 == 63)
		{
			(*gf) << std::endl;
		}
	}

	(*gf) << std::dec;
	(*gf) << std::endl;
	(*gf) << std::endl;

}

void PrintFitAllocatorUsedMap(TDAllocatorFit & a)
{
	(*gf) << "used memory map" << std::endl;

	char map[512];
	size_t i, isz;

	size_t sz;
	size_t ed;
	//size_t scale;
	void * start = a.GetMemoryStart();

	sz = (size_t)a.GetMemoryEnd() - (size_t)start;
	//scale = sz * sizeof(map);

	PDAllocatorFitChunk top;
	PDAllocatorFitChunk c;

	::memset(map, '0', sizeof(map));

	top = (PDAllocatorFitChunk)start;
	c = top;

	while(top)
	{
		size_t st;

		st = ((size_t)c->allocatedStart - (size_t)start) * sizeof(map) / sz;
		ed = ((size_t)c->freedStart - (size_t)start) * sizeof(map) / sz;

		::memset(&map[st], '1', ed - st + 1);
	
		c = c->next;
		if(c == top)
		{
			break;
		}
	}


	isz = sizeof(map);
	for(i = 0 ; i < isz ; i++)
	{
		(*gf) << map[i];
		if(i % 64 == 63)
		{
			(*gf) << std::endl;
		}
	}

	(*gf) << std::endl;
	(*gf) << std::endl;
}

void PrintFitAllocatorInfo(TDAllocatorFit & a)
{
	(*gf) << "Total Used Memory = " << a.GetTotalUsedSize() << std::endl;
	(*gf) << "Total Allocated Memory = " << a.GetTotalAllocatedSize() << std::endl;
	(*gf) << std::endl;
}

template<class TDContainer> void PrintClass1Container(TDContainer & v0)
{
	typename TDContainer::iterator i;
	typename TDContainer::iterator ied;

	ied = v0.end();
	for(i = v0.begin() ; i != ied ; i++)
	{
		(*gf) << "i->a = " << i->GetA() << " -> " << &*i << std::endl;
	}
	(*gf) << std::endl;
}

//
void STLAllocatorSample1_PlacementNew()
{
	std::cout << "STLAllocatorSample1_HowToUse" << std::endl;

	std::ofstream f("STLAllocatorSample1_PlacementNew.txt");

	gf = &f;

	//
	unsigned char d[256];
	::memset(d, 0xed, sizeof(d));
	TDAllocatorFit a(16, DycNextFit, false, d, d + sizeof(d));

	(*gf) << "MemoryStart = " << a.GetMemoryStart() << std::endl;
	(*gf) << "MemoryEnd = " << a.GetMemoryEnd() << std::endl;
	(*gf) << std::endl;

	PrintAllMemory(d, sizeof(d));
	PrintFitAllocatorUsedMap(a);
	PrintFitAllocatorInfo(a);


	//
	TDClass1 * c0;

	c0 = (TDClass1 *)a.Allocate(sizeof(TDClass1));
	new(c0) TDClass1();

	c0->SetA(15);
	(*gf) << "c0->a = " << c0->GetA() << std::endl;
	(*gf) << std::endl;

	PrintAllMemory(d, sizeof(d));
	PrintFitAllocatorUsedMap(a);
	PrintFitAllocatorInfo(a);

	c0->~TDClass1();
	a.Deallocate((void *)c0);

	PrintAllMemory(d, sizeof(d));
	PrintFitAllocatorInfo(a);

}

//
void STLAllocatorSample1_STLAllocation()
{
	std::cout << "STLAllocatorSample1_STLAllocation" << std::endl;

	std::ofstream f("STLAllocatorSample1_STLAllocation.txt");

	gf = &f;

	//
	unsigned char d[256];
	::memset(d, 0xed, sizeof(d));
	TDAllocatorFit a(16, DycNextFit, false, d, d + sizeof(d));
	typedef TDSTLAllocatorStatic<TDClass1> TDSA;
	TDSA sa(a);

	(*gf) << "MemoryStart = " << a.GetMemoryStart() << std::endl;
	(*gf) << "MemoryEnd = " << a.GetMemoryEnd() << std::endl;
	(*gf) << std::endl;

	PrintAllMemory(d, sizeof(d));
	PrintFitAllocatorInfo(a);

	//
	TDClass1 * c0;

	c0 = (TDClass1 *)sa.allocate(1, NULL);
	sa.construct((TDSA::pointer)c0, TDClass1());

	c0->SetA(15);
	(*gf) << "c0->a = " << c0->GetA() << std::endl;
	(*gf) << std::endl;

	PrintAllMemory(d, sizeof(d));
	PrintFitAllocatorUsedMap(a);
	PrintFitAllocatorInfo(a);

	sa.destroy(c0);
	sa.deallocate((TDSA::pointer)c0, 1);

	PrintAllMemory(d, sizeof(d));
	PrintFitAllocatorUsedMap(a);
	PrintFitAllocatorInfo(a);

}

//
void STLAllocatorSample1_STLVector()
{
	std::cout << "STLAllocatorSample1_STLVector" << std::endl;

	std::ofstream f("STLAllocatorSample1_STLVector.txt");

	gf = &f;

	//
	unsigned char d[256];
	::memset(d, 0xed, sizeof(d));
	TDAllocatorFit a(16, DycNextFit, false, d, d + sizeof(d));
	typedef TDSTLAllocatorStatic<TDClass1> TDSA;
	TDSA sa(a);

	(*gf) << "MemoryStart = " << a.GetMemoryStart() << std::endl;
	(*gf) << "MemoryEnd = " << a.GetMemoryEnd() << std::endl;
	(*gf) << std::endl;

	PrintAllMemory(d, sizeof(d));
	PrintFitAllocatorUsedMap(a);
	PrintFitAllocatorInfo(a);


	//
	typedef std::vector<TDClass1, TDSTLAllocatorStatic<TDClass1> > TDV;
	TDV v0(sa);
	TDV::iterator i;
	TDV::iterator ied;
	int ct;

	v0.resize(3);
	(*gf) << "&v0[0] = " << &v0[0] << std::endl;

	ct = 0;
	ied = v0.end();
	for(i = v0.begin() ; i != ied ; i++)
	{
		i->SetA(ct);
		ct++;
	}

	PrintClass1Container(v0);

	PrintAllMemory(d, sizeof(d));
	PrintFitAllocatorUsedMap(a);
	PrintFitAllocatorInfo(a);

}

//
void STLAllocatorSample1_STLVectorPushBack()
{
	std::cout << "STLAllocatorSample1_STLVectorPushBack" << std::endl;

	std::ofstream f("STLAllocatorSample1_STLVectorPushBack.txt");

	gf = &f;

	//
	unsigned char d[1024 * 12];
	::memset(d, 0xed, sizeof(d));
	TDAllocatorFit a(16, DycNextFit, false, d, d + sizeof(d));
	typedef TDSTLAllocatorStatic<TDClass1> TDSA;
	TDSA sa(a);

	(*gf) << "MemoryStart = " << a.GetMemoryStart() << std::endl;
	(*gf) << "MemoryEnd = " << a.GetMemoryEnd() << std::endl;
	(*gf) << std::endl;

	PrintAllMemory(d, sizeof(d));
	PrintFitAllocatorUsedMap(a);
	PrintFitAllocatorInfo(a);


	//
	typedef std::vector<TDClass1, TDSTLAllocatorStatic<TDClass1> > TDV;
	TDV v0(sa);

	size_t j, jsz;
	size_t oldCapacity = 0;

	jsz = 640;

	for(j = 0 ; j < jsz ; j++)
	{
		v0.push_back(TDClass1());
		v0.back().SetA(j);

		if(oldCapacity < v0.capacity())
		{
			PrintFitAllocatorUsedMap(a);

			oldCapacity = v0.capacity();
		}
	}

	PrintClass1Container(v0);

	PrintAllMemory(d, sizeof(d));
	PrintFitAllocatorUsedMap(a);
	PrintFitAllocatorInfo(a);

}

//
void STLAllocatorSample1_STLVectorPushBackReserved()
{
	std::cout << "STLAllocatorSample1_STLVectorPushBackReserved" << std::endl;

	std::ofstream f("STLAllocatorSample1_STLVectorPushBackReserved.txt");

	gf = &f;

	//
	unsigned char d[1024 * 12];
	::memset(d, 0xed, sizeof(d));
	TDAllocatorFit a(16, DycNextFit, false, d, d + sizeof(d));
	typedef TDSTLAllocatorStatic<TDClass1> TDSA;
	TDSA sa(a);

	(*gf) << "MemoryStart = " << a.GetMemoryStart() << std::endl;
	(*gf) << "MemoryEnd = " << a.GetMemoryEnd() << std::endl;
	(*gf) << std::endl;

	PrintAllMemory(d, sizeof(d));
	PrintFitAllocatorUsedMap(a);
	PrintFitAllocatorInfo(a);


	//
	typedef std::vector<TDClass1, TDSTLAllocatorStatic<TDClass1> > TDV;
	TDV v0(sa);

	size_t j, jsz;
	size_t oldCapacity = 0;

	v0.reserve((sizeof(d) - a.GetHeaderSize() - sizeof(TDAllocatorFitChunk) - 1) / sizeof(TDClass1));

	jsz = v0.capacity();

	for(j = 0 ; j < jsz ; j++)
	{
		v0.push_back(TDClass1());
		v0.back().SetA(j);

		if(oldCapacity < v0.capacity())
		{
			PrintFitAllocatorUsedMap(a);

			oldCapacity = v0.capacity();
		}
	}

	PrintClass1Container(v0);

	PrintAllMemory(d, sizeof(d));
	PrintFitAllocatorUsedMap(a);
	PrintFitAllocatorInfo(a);

}

//
void STLAllocatorSample1_STLList()
{
	std::cout << "STLAllocatorSample1_STLList" << std::endl;

	std::ofstream f("STLAllocatorSample1_STLList.txt");

	gf = &f;

	//
	unsigned char d[256];
	::memset(d, 0xed, sizeof(d));
	TDAllocatorFit a(16, DycNextFit, false, d, d + sizeof(d));
	typedef TDSTLAllocatorStatic<TDClass1> TDSA;
	TDSA sa(a);

	(*gf) << "MemoryStart = " << a.GetMemoryStart() << std::endl;
	(*gf) << "MemoryEnd = " << a.GetMemoryEnd() << std::endl;
	(*gf) << std::endl;

	PrintAllMemory(d, sizeof(d));
	PrintFitAllocatorUsedMap(a);
	PrintFitAllocatorInfo(a);


	//
	typedef std::list<TDClass1, TDSTLAllocatorStatic<TDClass1> > TDV;
	TDV v0(sa);
	TDV::iterator i;
	TDV::iterator ied;
	int ct;

	v0.resize(3);

	ct = 0;
	ied = v0.end();
	for(i = v0.begin() ; i != ied ; i++)
	{
		i->SetA(ct);
		ct++;
	}

	PrintClass1Container(v0);

	PrintAllMemory(d, sizeof(d));
	PrintFitAllocatorUsedMap(a);
	PrintFitAllocatorInfo(a);

}

//
void STLAllocatorSample1_STLListSplice()
{
	std::cout << "STLAllocatorSample1_STLListSplice" << std::endl;

	std::ofstream f("STLAllocatorSample1_STLListSplice.txt");

	gf = &f;

	//
	unsigned char d[512];
	::memset(d, 0xed, sizeof(d));
	TDAllocatorFit a(16, DycNextFit, false, d, d + sizeof(d));
	typedef TDSTLAllocatorStatic<TDClass1> TDSA;
	TDSA sa(a);

	(*gf) << "MemoryStart = " << a.GetMemoryStart() << std::endl;
	(*gf) << "MemoryEnd = " << a.GetMemoryEnd() << std::endl;
	(*gf) << std::endl;

	PrintAllMemory(d, sizeof(d));
	PrintFitAllocatorUsedMap(a);
	PrintFitAllocatorInfo(a);


	//
	typedef std::list<TDClass1, TDSTLAllocatorStatic<TDClass1> > TDV;
	TDV v0(sa);
	TDV v1(sa);
	TDV::iterator i;
	TDV::iterator ied;
	int ct;

	v0.resize(3);
	ct = 0;
	ied = v0.end();
	for(i = v0.begin() ; i != ied ; i++)
	{
		i->SetA(ct);
		ct++;
	}

	v1.resize(3);
	ied = v1.end();
	for(i = v1.begin() ; i != ied ; i++)
	{
		i->SetA(ct);
		ct++;
	}

	(*gf) << "v0" << std::endl;
	PrintClass1Container(v0);

	(*gf) << "v1" << std::endl;
	PrintClass1Container(v1);

	//
	i = v0.begin();
	i++;
	v0.splice(i, v1);

	//
	(*gf) << "v0" << std::endl;
	PrintClass1Container(v0);

	(*gf) << "v1" << std::endl;
	PrintClass1Container(v1);


	PrintAllMemory(d, sizeof(d));
	PrintFitAllocatorUsedMap(a);
	PrintFitAllocatorInfo(a);

}

//
void STLAllocatorSample1_STLListSpliceStatefull()
{
	std::cout << "STLAllocatorSample1_STLListSpliceStatefull" << std::endl;

	std::ofstream f("STLAllocatorSample1_STLListSpliceStatefull.txt");

	gf = &f;

	try
	{
		//
		unsigned char d0[512];
		unsigned char d1[512];
		::memset(d0, 0xed, sizeof(d0));
		::memset(d1, 0xed, sizeof(d1));
		TDAllocatorFit a0(16, DycNextFit, false, d0, d0 + sizeof(d0));
		TDAllocatorFit a1(16, DycNextFit, false, d1, d1 + sizeof(d1));
		typedef TDSTLAllocatorStatefull<TDClass1> TDSA;
		TDSA sa0(a0);
		TDSA sa1(a1);

		(*gf) << "a0 MemoryStart = " << a0.GetMemoryStart() << std::endl;
		(*gf) << "a0 MemoryEnd = " << a0.GetMemoryEnd() << std::endl;
		(*gf) << std::endl;

		(*gf) << "a1 MemoryStart = " << a1.GetMemoryStart() << std::endl;
		(*gf) << "a1 MemoryEnd = " << a1.GetMemoryEnd() << std::endl;
		(*gf) << std::endl;

		(*gf) << "d0" << std::endl;
		PrintAllMemory(d0, sizeof(d0));
		PrintFitAllocatorUsedMap(a0);
		PrintFitAllocatorInfo(a0);
		(*gf) << std::endl;

		(*gf) << "d1" << std::endl;
		PrintAllMemory(d1, sizeof(d1));
		PrintFitAllocatorUsedMap(a1);
		PrintFitAllocatorInfo(a1);
		(*gf) << std::endl;
		

		//
		typedef std::list<TDClass1, TDSA > TDV;
		TDV v0(sa0);
		TDV v1(sa1);
		TDV::iterator i;
		TDV::iterator ied;
		int ct;

		v0.resize(3);
		ct = 0;
		ied = v0.end();
		for(i = v0.begin() ; i != ied ; i++)
		{
			i->SetA(ct);
			ct++;
		}

		v1.resize(3);
		ied = v1.end();
		for(i = v1.begin() ; i != ied ; i++)
		{
			i->SetA(ct);
			ct++;
		}

		(*gf) << "v0" << std::endl;
		PrintClass1Container(v0);

		(*gf) << "v1" << std::endl;
		PrintClass1Container(v1);

		//
		i = v0.begin();
		i++;
		v0.splice(i, v1);

		//
		(*gf) << "v0" << std::endl;
		PrintClass1Container(v0);

		(*gf) << "v1" << std::endl;
		PrintClass1Container(v1);


		(*gf) << "d0" << std::endl;
		PrintAllMemory(d0, sizeof(d0));
		PrintFitAllocatorUsedMap(a0);
		PrintFitAllocatorInfo(a0);
		(*gf) << std::endl;

		(*gf) << "d1" << std::endl;
		PrintAllMemory(d1, sizeof(d1));
		PrintFitAllocatorUsedMap(a1);
		PrintFitAllocatorInfo(a1);
		(*gf) << std::endl;
	}
	catch(...)
	{
		(*gf) << "exception called." << std::endl;
		(*gf) << "maybe cannot use stateful allocator." << std::endl;
	}
}

//

void STLAllocatorSample1_Fragmentation(unsigned int fitType, bool antiFragmentation)
{
	unsigned char d[1024 * 16];
	::memset(d, 0xed, sizeof(d));
	TDAllocatorFit a(16, fitType, antiFragmentation, d, d + sizeof(d));

	(*gf) << "MemoryStart = " << a.GetMemoryStart() << std::endl;
	(*gf) << "MemoryEnd = " << a.GetMemoryEnd() << std::endl;
	(*gf) << std::endl;

	PrintAllMemory(d, sizeof(d));
	PrintFitAllocatorUsedMap(a);
	PrintFitAllocatorInfo(a);

	const size_t sz = 1024;
	void * p[sz * 2];
	size_t i, isz;

	isz = sz;

	try
	{
		for(i = 0 ; i < isz ; i++)
		{
			p[i * 2 + 0] = a.Allocate(i + 1);
			p[i * 2 + 1] = a.Allocate(1);
			a.Deallocate(p[i * 2 + 0]);
		}
	}
	catch(...)
	{
		(*gf) << "exception called." << std::endl;
		(*gf) << "maybe cannot allocate." << std::endl;
	}

	PrintAllMemory(d, sizeof(d));
	PrintFitAllocatorUsedMap(a);
	PrintFitAllocatorInfo(a);

	(*gf) << "iteration count = " << i << std::endl;

}

void STLAllocatorSample1_Fragmentation2(unsigned int fitType, bool antiFragmentation)
{
	unsigned char d[1024 * 16];
	::memset(d, 0xed, sizeof(d));
	TDAllocatorFit a(16, fitType, antiFragmentation, d, d + sizeof(d));

	(*gf) << "MemoryStart = " << a.GetMemoryStart() << std::endl;
	(*gf) << "MemoryEnd = " << a.GetMemoryEnd() << std::endl;
	(*gf) << std::endl;

	PrintAllMemory(d, sizeof(d));
	PrintFitAllocatorUsedMap(a);
	PrintFitAllocatorInfo(a);

	const size_t sz = 1024;
	void * p[sz * 2];
	size_t i, isz;

	isz = sz;

	try
	{
		for(i = 0 ; i < isz ; i++)
		{
			p[i * 2 + 0] = a.Allocate(i + 1);
			p[i * 2 + 1] = a.Allocate(i + 1);
			a.Deallocate(p[i * 2 + 0]);
		}
	}
	catch(...)
	{
		(*gf) << "exception called." << std::endl;
		(*gf) << "maybe cannot allocate." << std::endl;
	}

	PrintAllMemory(d, sizeof(d));
	PrintFitAllocatorUsedMap(a);
	PrintFitAllocatorInfo(a);

	(*gf) << "iteration count = " << i << std::endl;

}

void STLAllocatorSample1_NoFragmentation(unsigned int fitType, bool antiFragmentation)
{
	unsigned char d[1024 * 16];
	::memset(d, 0xed, sizeof(d));
	TDAllocatorFit a(16, fitType, antiFragmentation, d, d + sizeof(d));

	(*gf) << "MemoryStart = " << a.GetMemoryStart() << std::endl;
	(*gf) << "MemoryEnd = " << a.GetMemoryEnd() << std::endl;
	(*gf) << std::endl;

	PrintAllMemory(d, sizeof(d));
	PrintFitAllocatorUsedMap(a);
	PrintFitAllocatorInfo(a);

	const size_t sz = 1024;
	void * p[sz * 2];
	size_t i, isz;

	isz = sz;

	try
	{
		for(i = 0 ; i < isz ; i++)
		{
			p[i * 2 + 0] = a.Allocate(i + 1);
			p[i * 2 + 1] = a.Allocate(1);
		}
	}
	catch(...)
	{
		(*gf) << "exception called." << std::endl;
		(*gf) << "maybe cannot allocate." << std::endl;
	}

	PrintAllMemory(d, sizeof(d));
	PrintFitAllocatorUsedMap(a);
	PrintFitAllocatorInfo(a);

	(*gf) << "iteration count = " << i << std::endl;

}

void STLAllocatorSample1_NoFragmentation2(unsigned int fitType, bool antiFragmentation)
{
	unsigned char d[1024 * 16];
	::memset(d, 0xed, sizeof(d));
	TDAllocatorFit a(16, fitType, antiFragmentation, d, d + sizeof(d));

	(*gf) << "MemoryStart = " << a.GetMemoryStart() << std::endl;
	(*gf) << "MemoryEnd = " << a.GetMemoryEnd() << std::endl;
	(*gf) << std::endl;

	PrintAllMemory(d, sizeof(d));
	PrintFitAllocatorUsedMap(a);
	PrintFitAllocatorInfo(a);

	const size_t sz = 1024;
	void * p[sz * 2];
	size_t i, isz;

	isz = sz;

	try
	{
		for(i = 0 ; i < isz ; i++)
		{
			p[i * 2 + 0] = a.Allocate(i + 1);
			p[i * 2 + 1] = a.Allocate(i + 1);
		}
	}
	catch(...)
	{
		(*gf) << "exception called." << std::endl;
		(*gf) << "maybe cannot allocate." << std::endl;
	}

	PrintAllMemory(d, sizeof(d));
	PrintFitAllocatorUsedMap(a);
	PrintFitAllocatorInfo(a);

	(*gf) << "iteration count = " << i << std::endl;

}

void STLAllocatorSample1_Fragmentation()
{
	std::cout << "STLAllocatorSample1_Fragmentation" << std::endl;

	{
		std::ofstream f("STLAllocatorSample1_Fragmentation_NextFit_Frag.txt");
		gf = &f;

		STLAllocatorSample1_Fragmentation(DycNextFit, false);
	}

	{
		std::ofstream f("STLAllocatorSample1_Fragmentation_NextFit_AntiFrag.txt");
		gf = &f;

		STLAllocatorSample1_Fragmentation(DycNextFit, true);
	}

	{
		std::ofstream f("STLAllocatorSample1_Fragmentation_FirstFit_Frag.txt");
		gf = &f;

		STLAllocatorSample1_Fragmentation(DycFirstFit, false);
	}

	{
		std::ofstream f("STLAllocatorSample1_Fragmentation_FirstFit_AntiFrag.txt");
		gf = &f;

		STLAllocatorSample1_Fragmentation(DycFirstFit, true);
	}

	{
		std::ofstream f("STLAllocatorSample1_Fragmentation2_NextFit_Frag.txt");
		gf = &f;

		STLAllocatorSample1_Fragmentation2(DycNextFit, false);
	}

	{
		std::ofstream f("STLAllocatorSample1_Fragmentation2_NextFit_AntiFrag.txt");
		gf = &f;

		STLAllocatorSample1_Fragmentation2(DycNextFit, true);
	}

	{
		std::ofstream f("STLAllocatorSample1_Fragmentation2_FirstFit_Frag.txt");
		gf = &f;

		STLAllocatorSample1_Fragmentation2(DycFirstFit, false);
	}

	{
		std::ofstream f("STLAllocatorSample1_Fragmentation2_FirstFit_AntiFrag.txt");
		gf = &f;

		STLAllocatorSample1_Fragmentation2(DycFirstFit, true);
	}

	//

	{
		std::ofstream f("STLAllocatorSample1_NoFragmentation_NextFit_Frag.txt");
		gf = &f;

		STLAllocatorSample1_NoFragmentation(DycNextFit, false);
	}

	{
		std::ofstream f("STLAllocatorSample1_NoFragmentation_NextFit_AntiFrag.txt");
		gf = &f;

		STLAllocatorSample1_NoFragmentation(DycNextFit, true);
	}

	{
		std::ofstream f("STLAllocatorSample1_NoFragmentation_FirstFit_Frag.txt");
		gf = &f;

		STLAllocatorSample1_NoFragmentation(DycFirstFit, false);
	}

	{
		std::ofstream f("STLAllocatorSample1_NoFragmentation_FirstFit_AntiFrag.txt");
		gf = &f;

		STLAllocatorSample1_NoFragmentation(DycFirstFit, true);
	}

	{
		std::ofstream f("STLAllocatorSample1_NoFragmentation2_NextFit_Frag.txt");
		gf = &f;

		STLAllocatorSample1_NoFragmentation2(DycNextFit, false);
	}

	{
		std::ofstream f("STLAllocatorSample1_NoFragmentation2_NextFit_AntiFrag.txt");
		gf = &f;

		STLAllocatorSample1_NoFragmentation2(DycNextFit, true);
	}

	{
		std::ofstream f("STLAllocatorSample1_NoFragmentation2_FirstFit_Frag.txt");
		gf = &f;

		STLAllocatorSample1_NoFragmentation2(DycFirstFit, false);
	}

	{
		std::ofstream f("STLAllocatorSample1_NoFragmentation2_FirstFit_AntiFrag.txt");
		gf = &f;

		STLAllocatorSample1_NoFragmentation2(DycFirstFit, true);
	}

	
}

//
template<class TDType> class TDVec3
{
public:
	TDType x;
	TDType y;
	TDType z;

	TDVec3()
	{
	
	}

	TDVec3(TDType _x, TDType _y, TDType _z)
	{
		x = _x;
		y = _y;
		z = _z;
	}

	void Print(std::ofstream & f)
	{
		f << "(" << x << ", " << y << ", " << z << "), ";
	}

};

template<class TDType>void TDVec3Print(std::ofstream & f, TDType & d)
{
	typename TDType::iterator i, ied;

	ied = d.end();
	for(i = d.begin() ; i != ied ; i++)
	{
		i->Print(f);
	}

	f << std::endl;
}

template<class TDType> class TDClass2
{
public:
	struct TVA : public TDVec3<TDType>
	{
		TVA(){}

		TVA(TDType _x, TDType _y, TDType _z) : TDVec3<TDType>(_x, _y, _z) {}

	};

	struct TVB : public TDVec3<TDType>
	{
		TVB(){}

		TVB(TDType _x, TDType _y, TDType _z) : TDVec3<TDType>(_x, _y, _z) {}

	};

	typedef TDSTLAllocatorStatic<TVA> TVAA;
	typedef TDSTLAllocatorStatic<TVB> TVBA;

	typedef std::vector<TVA, TVAA> TVAV;
	typedef std::list<TVB, TVBA> TVBV;

	TVAV v0;
	TVBV v1;

	TDClass2(TVAA a0, TVBA a1) : v0(a0), v1(a1)
	{
		
	}

	void Init()
	{
		v0.resize(4);
		v0[0] = TVA(0.0, 1.0, 2.0);
		v0[1] = TVA(3.0, 4.0, 5.0);
		v0[2] = TVA(6.0, 7.0, 8.0);
		v0[3] = TVA(9.0, 10.0, 11.0);

		v1.push_back(TVB(12.0, 13.0, 14.0));
		v1.push_back(TVB(15.0, 16.0, 17.0));
		v1.push_back(TVB(18.0, 19.0, 20.0));
		v1.push_back(TVB(21.0, 22.0, 23.0));

	}

	void Print(std::ofstream & f)
	{
		f << "v0" << std::endl;
		TDVec3Print(f, v0);

		f << "v1" << std::endl;
		TDVec3Print(f, v1);
	}

};

void STLAllocatorSample1_TypeAllocator()
{
	std::cout << "STLAllocatorSample1_TypeAllocator" << std::endl;

	std::ofstream f("STLAllocatorSample1_TypeAllocator.txt");
	gf = &f;

	typedef TDClass2<float> TDC;

	unsigned char d0[512];
	unsigned char d1[512];
	::memset(d0, 0xed, sizeof(d0));
	::memset(d1, 0xed, sizeof(d1));
	TDAllocatorFit a0(16, DycNextFit, false, d0, d0 + sizeof(d0));
	TDAllocatorFit a1(16, DycNextFit, false, d1, d1 + sizeof(d1));
	typedef TDSTLAllocatorStatefull<TDClass1> TDSA;
	TDC::TVAA sa0(a0);
	TDC::TVBA sa1(a1);

	(*gf) << "a0 MemoryStart = " << a0.GetMemoryStart() << std::endl;
	(*gf) << "a0 MemoryEnd = " << a0.GetMemoryEnd() << std::endl;
	(*gf) << std::endl;

	(*gf) << "a1 MemoryStart = " << a1.GetMemoryStart() << std::endl;
	(*gf) << "a1 MemoryEnd = " << a1.GetMemoryEnd() << std::endl;
	(*gf) << std::endl;

	(*gf) << "d0" << std::endl;
	PrintAllMemory(d0, sizeof(d0));
	PrintFitAllocatorUsedMap(a0);
	PrintFitAllocatorInfo(a0);
	(*gf) << std::endl;

	(*gf) << "d1" << std::endl;
	PrintAllMemory(d1, sizeof(d1));
	PrintFitAllocatorUsedMap(a1);
	PrintFitAllocatorInfo(a1);
	(*gf) << std::endl;

	//
	TDC c(sa0, sa1);

	c.Init();

	c.Print((*gf));

	(*gf) << "d0" << std::endl;
	PrintAllMemory(d0, sizeof(d0));
	PrintFitAllocatorUsedMap(a0);
	PrintFitAllocatorInfo(a0);
	(*gf) << std::endl;

	(*gf) << "d1" << std::endl;
	PrintAllMemory(d1, sizeof(d1));
	PrintFitAllocatorUsedMap(a1);
	PrintFitAllocatorInfo(a1);
	(*gf) << std::endl;

}


//
void STLAllocatorSample1()
{
	STLAllocatorSample1_PlacementNew();
	STLAllocatorSample1_STLAllocation();
	STLAllocatorSample1_STLVector();
	STLAllocatorSample1_STLVectorPushBack();
	STLAllocatorSample1_STLVectorPushBackReserved();
	STLAllocatorSample1_STLList();
	STLAllocatorSample1_STLListSplice();
	STLAllocatorSample1_STLListSpliceStatefull();
	STLAllocatorSample1_Fragmentation();
	STLAllocatorSample1_TypeAllocator();
}

