#include "TDSTLAllocatorSample.h"

// new delete allocator

void TDAllocatorNew::SetAlignment(size_t alignment)
{
	this->alignment = alignment;
}

size_t TDAllocatorNew::GetAlignment()
{
	return alignment;
}

void * TDAllocatorNew::Allocate(size_t size)
{
	unsigned char * tp;

	tp = new unsigned char[size + alignment + sizeof(void *)];

	count++;

	return AlignedPointer((void *)tp, alignment);
}

void TDAllocatorNew::Deallocate(void * p)
{
	p = UnAlignedPointer(p);
	delete [] (unsigned char *)p;

	count--;
}

bool TDAllocatorNew::IsEmpty()
{
	if(count == 0)
	{
		return true;
	}

	return false;
}

// linked list allocator

void TDAllocatorFitChunk::Init()
{
	next = this;
	prev = this;
}

void TDAllocatorFitChunk::Remove()
{
	next->prev = prev;
	prev->next = next;
	next = this;
	prev = this;
}

void TDAllocatorFitChunk::InsertNext(PDAllocatorFitChunk nxt)
{
	nxt->next->prev = nxt->prev;
	nxt->prev->next = nxt->next;

	nxt->prev = this;
	nxt->next = next;
	next->prev = nxt;
	next = nxt;
}


TDAllocatorFit::~TDAllocatorFit()
{
	
}

void TDAllocatorFit::MemInit()
{
	PDAllocatorFitChunk top = (PDAllocatorFitChunk)memoryStart;
	currentChunk = top;

	top->Init();
	top->allocatedStart = memoryStart;
	top->freedStart = memoryStart + sizeof(TDAllocatorFitChunk);
	top->cunkEnd = memoryEnd;

	totalUsedSize = sizeof(TDAllocatorFitChunk);
	totalAllocatedSize = 0;
}

void * TDAllocatorFit::GetMemoryStart()
{
	return (void *)memoryStart;
}

void * TDAllocatorFit::GetMemoryEnd()
{
	return (void *)memoryEnd;
}

void * TDAllocatorFit::Allocate(size_t size)
{

	size_t sz;
	ptrdiff_t frsz;
	sz = size + GetHeaderSize();

	if(antiFragmentation)
	{
		size_t i;

		i = 1;
		while(size > i)
		{
			i = i << 1;
		}

		sz = i + GetHeaderSize();
	}

	PDAllocatorFitChunk top = currentChunk;

	if(fitType == DycNextFit)
	{
		top = currentChunk;
	}
	else if(fitType == DycFirstFit)
	{
		currentChunk = (PDAllocatorFitChunk)memoryStart;
		top = (PDAllocatorFitChunk)memoryStart;
	}

	while(top)
	{
		frsz = (ptrdiff_t)top->cunkEnd - (ptrdiff_t)top->freedStart;
		if(frsz > (ptrdiff_t)sz)
		{
			break;
		}
	
		top = top->next;
		if(top == currentChunk)
		{
			throw std::bad_alloc();
			return NULL;
		}
	}

	currentChunk = top;

	unsigned char * _p;
	PDAllocatorFitChunk nxt;

	_p = top->freedStart;

	::memset(top->freedStart, 0xac, sz);

	nxt = (PDAllocatorFitChunk)top->freedStart;
	nxt->Init();
	nxt->allocatedStart = top->freedStart;
	nxt->freedStart = nxt->allocatedStart + sz;
	nxt->cunkEnd = top->cunkEnd;
	nxt->allocatedSize = size;
	top->cunkEnd = nxt->allocatedStart;

	top->InsertNext(nxt);

	//
	unsigned char * cp;
	unsigned char * alcp;
	void * ppt;

	cp = (unsigned char *)_p;
	alcp = (unsigned char *)(((size_t)(cp + GetHeaderSize() - 1) / alignment) * alignment);
	ppt = (void *)(alcp - sizeof(void *));

	*(void **)ppt = (void *)_p;

	totalUsedSize += sz;
	totalAllocatedSize += size;

	return (void *)alcp;
}

void TDAllocatorFit::Deallocate(void * p)
{
	unsigned char * cp;
	cp = (unsigned char *)p;

	if((size_t)p < (size_t)memoryStart || (size_t)memoryEnd <= (size_t)p)
	{
		throw std::bad_exception();
	}

	cp = cp - sizeof(void *);
	
	PDAllocatorFitChunk top = (PDAllocatorFitChunk)*(void * *)cp;
	PDAllocatorFitChunk prev = top->prev;
	
	if((size_t)top <= (size_t)prev)
	{
		throw std::bad_exception();
	}

	prev->cunkEnd = top->cunkEnd;

	if((size_t)prev < (size_t)currentChunk)
	{
		currentChunk = prev;
	}

	top->Remove();

	totalUsedSize -= (size_t)top->freedStart - (size_t)top->allocatedStart;
	totalAllocatedSize -= top->allocatedSize;

	::memset(top, 0xed, (size_t)top->freedStart - (size_t)top->allocatedStart);
}

size_t TDAllocatorFit::GetTotalUsedSize()
{
	return totalUsedSize;
}

size_t TDAllocatorFit::GetTotalAllocatedSize()
{
	return totalAllocatedSize;
}

size_t TDAllocatorFit::GetHeaderSize()
{
	return alignment + sizeof(TDAllocatorFitChunk) + sizeof(void *);
}

bool TDAllocatorFit::IsEmpty()
{
	PDAllocatorFitChunk top = (PDAllocatorFitChunk)memoryStart;

	if(top->next == top)
	{
		return true;
	}

	return false;
}
